﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading.Tasks;
using ClientServer.Controllers.Core.Exercises;
using ClientServer.Db;
using ClientServer.Helpers;
using ClientServer.Models.CustomProjects;
using ClientServer.Models.Exercises;
using ClientServer.Models.Exercises.Solution;
using ClientServer.Models.Exercises.Tests;
using ClientServer.Models.Files;
using Microsoft.AspNetCore.Http;
using Microsoft.AspNetCore.Mvc;
using Microsoft.EntityFrameworkCore;
using Microsoft.Extensions.Primitives;

namespace ClientServer.Controllers.Core.customProjects
{
    [Route(Constants.ApiPrefix + "custom/projects/edit")]
    public class EditCustomProjectController : ControllerWithDb
    {
        public EditCustomProjectController(SyndromDbContext context) : base(context)
        {
        }

        [HttpGet("{customProjectId}")]
        public async Task GetOwnCustomProject(int customProjectId)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();

            var customProject = await _context.CustomProjects.Where(p =>
                        p.UserId == userId && //just ensure the project is from the current user
                        p.Id == customProjectId
                    )
                    .Include(p => p.Description)
                    .ThenInclude(p => p.AssetReferences)
                    .ThenInclude(p => p.FileReferenceUserFileAsset)
                    .Include(p => p.Tests)
                    .ThenInclude(p => p.AssetReferences)
                    .ThenInclude(p => p.FileReferenceUserFileAsset)
                    .Include(p => p.Tests)
                    .ThenInclude(p => p.TestSettings)
                    .Include(p => p.Tests)
                    .FirstOrDefaultAsync()
                ;

            if (customProject == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "custom project not found")));
                return;
            }

            var lastEditorSolution = await _context.CustomProjectSolutions.Where(p =>
                    p.CustomProjectId == customProject.Id &&
                    p.PLangId == customProject.LastEditorPLangId
                )
                .Include(p => p.SolutionFiles)
                .FirstOrDefaultAsync();

            if (lastEditorSolution == null)
            {
                //but we could get some other plangs solution... but maybe thats not helpful
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "no solution found")));
                return;
            }

            var markdownUserFilesIdsToRead = customProject.Description.AssetReferences
                .Select(p => p.FileReferenceUserFileAssetId)
                .ToArray();

            var markdownUserAssetsBasePath = Files.GetUploadFilePath(UploadDirType.UserAssets);

            var markdownUserFilesDict =
                await Files.ReadUploadedFiles(markdownUserAssetsBasePath, markdownUserFilesIdsToRead);


            var customProjectForFrontend = new CustomProjectFullBase()
            {
                Id = customProject.Id,
                DisplayName = customProject.DisplayName,
                LastEditorPLangId = customProject.LastEditorPLangId,
                Description = new CustomProjectDescriptionFullBase()
                {
                    Id = customProject.Description.Id,
                    Content = customProject.Description.Content,
                    Assets = customProject.Description.AssetReferences.Select(p => new FileWithData()
                    {
                        Id = p.FileReferenceUserFileAsset.Id,
                        DisplayName = p.FileReferenceUserFileAsset.OriginalName,
                        MimeType = p.FileReferenceUserFileAsset.MimeType,
                        Content = markdownUserFilesDict[p.FileReferenceUserFileAssetId]
                    }).ToList()
                },
                Solution = new CustomProjectSolutionFullBase()
                {
                    Id = lastEditorSolution.Id,
                    CreatedAt = lastEditorSolution.CreatedAt,
                    MainFileId = lastEditorSolution.MainFileId ?? 0,
                    PLangId = lastEditorSolution.PLangId,
                    SolutionFiles = lastEditorSolution.SolutionFiles.Select(p => new CustomProjectSolutionFileFullBase()
                    {
                        Id = p.Id,
                        Content = p.Content,
                        DisplayIndex = p.DisplayIndex,
                        FileNameWithExtension = p.FileNameWithExtension,
                        UiIsDisplayed = p.IsDisplayed
                    }).ToList()
                },
                Tests = customProject.Tests.Select(p => new CustomProjectTestFullBase()
                {
                    Id = p.Id,
                    Content = p.Content,
                    DisplayIndex = p.DisplayIndex,
                    DisplayName = p.DisplayName,
                    TestTypeId = p.TestTypeId,
                    Weight = p.Weight,
                    Files = p.AssetReferences.Select(asset => new FilePreviewFromBackend()
                    {
                        Id = asset.FileReferenceUserFileAsset.Id,
                        OriginalName = asset.FileReferenceUserFileAsset.OriginalName,
                        MimeType = asset.FileReferenceUserFileAsset.MimeType,
                        SizeInBytes = asset.FileReferenceUserFileAsset.SizeInBytes,
                    }).ToList(),
                    TestSettings = new TestSettingsFullBase()
                    {
                        Id = p.TestSettings.Id,
                        MaxDiskSpaceInKb = p.TestSettings.MaxDiskSpaceInKb,
                        MemoryLimitInKb = p.TestSettings.MemoryLimitInKb,
                        TimeoutInMs = p.TestSettings.TimeoutInMs,
                        CompileTimeoutInMs = p.TestSettings.CompileTimeoutInMs,
                        CompilerOptions = p.TestSettings.CompilerOptions,
                    }
                }).ToList()
            };

            await
                Response.WriteAsync(
                    Jc.Serialize(
                        new BasicResponseWithData<CustomProjectFullBase>(ResponseCode.Ok, "",
                            customProjectForFrontend)));
        }


        /// <summary>
        /// returns all custom test results for the given own custom project
        /// </summary>
        /// <param name="customProjectId"></param>
        /// <param name="plangId"></param>
        [HttpGet("{customProjectId}/testresults/{plangId}")]
        public async Task GetOwnCustomProjectTestResults(int customProjectId, int plangId)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();

            var customProject = await _context.CustomProjects
                    .FirstOrDefaultAsync(p => p.UserId == userId && //just ensure the project is from the current user
                                              p.Id == customProjectId)
                ;

            if (customProject == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "custom project not found")));
                return;
            }


            var pLang = await _context.PLangs.FirstOrDefaultAsync(p => p.Id == plangId);

            if (pLang == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "plang not found")));
                return;
            }

            var solution = await _context.CustomProjectSolutions
                    .Include(p => p.TestResults)
                    .FirstOrDefaultAsync(p => p.CustomProjectId == customProjectId && p.PLangId == pLang.Id)
                ;

            if (solution == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "solution not found")));
                return;
            }

            var storedTestResults = solution.TestResults.Select(p => new StoredTestResultFromBackend()
                {
                    TestId = p.CustomProjectTestId,
                    Passed = p.Passed,
                    ProgramExitCode = p.ProgramExitCode,
                    TestResultCode = p.TestResultCode,
                    Protocol =
                        p.Protocol.Split(new string[] {Constants.TestProtocolSplitString},
                            StringSplitOptions.RemoveEmptyEntries).ToList(),
                    HasCompiled = p.HasCompiled
                }).ToList()
                ;

            await
                Response.WriteAsync(
                    Jc.Serialize(new BasicResponseWithData<List<StoredTestResultFromBackend>>(ResponseCode.Ok,
                        "", storedTestResults)));
        }

        [HttpPost("create")]
        public async Task CreateCustomProject([FromBody] CustomProjectFullBase customProjectFromFrontend)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();

            var systemSettings = await _context.SystemSettings.FirstOrDefaultAsync();

            if (systemSettings == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "system settings not found")));
                return;
            }

            var countProjects = await _context.CustomProjects.CountAsync(p => p.UserId == userId);

            //systemSettings.MaxCustomProjectsPerUser < 0 is unlimited
            if (systemSettings.MaxCustomProjectsPerUser == 0 ||
                (countProjects + 1 >= systemSettings.MaxCustomProjectsPerUser &&
                 systemSettings.MaxCustomProjectsPerUser > 0))
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest,
                            "max amount of custom projects exceeded, max amount is " +
                            systemSettings.MaxCustomProjectsPerUser)));
                return;
            }

            if (customProjectFromFrontend == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest,
                            "no project received (probably no solution/files created)")));
                return;
            }

            if (customProjectFromFrontend.Solution == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest,
                            "no solution found")));
                return;
            }

            var lastEditedPLang =
                await _context.PLangs.FirstOrDefaultAsync(p => p.Id == customProjectFromFrontend.LastEditorPLangId);


            if (lastEditedPLang == null)
            {
                var pLang = await _context.PLangs.FirstOrDefaultAsync(p =>
                    p.Id == customProjectFromFrontend.Solution.PLangId);

                if (pLang == null)
                {
                    await
                        Response.WriteAsync(
                            Jc.Serialize(new BasicResponse(ResponseCode.NotFound,
                                "last edited plang or plang not found")));
                    return;
                }

                lastEditedPLang = pLang;
            }


            #region tests

            //assets are managed separately

            var allTestTypes = await _context.TestTypes.ToListAsync();

            var tests = new List<CustomProjectTest>();

            foreach (var test in customProjectFromFrontend.Tests)
            {
                var testType = allTestTypes.FirstOrDefault(p => p.Id == test.TestTypeId);

                if (testType == null)
                {
                    await
                        Response.WriteAsync(
                            Jc.Serialize(new BasicResponse(ResponseCode.NotFound,
                                $"test type id {test.TestTypeId} was not found for test '{test.DisplayName}'")));
                    return;
                }

                tests.Add(new CustomProjectTest()
                {
                    Content = test.Content,
                    DisplayIndex = test.DisplayIndex,
                    DisplayName = test.DisplayName,
                    TestTypeId = testType.Id,
                    Weight = test.Weight,
                    TestSettings = new TestSettings()
                    {
                        MemoryLimitInKb = test.TestSettings.MemoryLimitInKb,
                        TimeoutInMs = test.TestSettings.TimeoutInMs,
                        MaxDiskSpaceInKb = test.TestSettings.MaxDiskSpaceInKb,
                        CompileTimeoutInMs = test.TestSettings.CompileTimeoutInMs,
                        CompilerOptions = test.TestSettings.CompilerOptions,
                    },
                });
            }

            #endregion

            #region first solution

            if (customProjectFromFrontend.Solution.SolutionFiles.Any(p =>
                p.Id == customProjectFromFrontend.Solution.MainFileId))
            {
                //ok
            }
            else
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound,
                            "main file not found")));
                return;
            }

            var solutionFiles = new List<CustomProjectSolutionFile>();
            CustomProjectSolutionFile mainFile = null;

            foreach (var solutionFileFromFrontend in customProjectFromFrontend.Solution.SolutionFiles)
            {
                var solutionFile = new CustomProjectSolutionFile()
                {
                    Content = solutionFileFromFrontend.Content,
                    DisplayIndex = solutionFileFromFrontend.DisplayIndex,
                    FileNameWithExtension = solutionFileFromFrontend.FileNameWithExtension,
                    IsDisplayed = solutionFileFromFrontend.UiIsDisplayed
                };

                if (solutionFileFromFrontend.Id == customProjectFromFrontend.Solution.MainFileId)
                {
                    mainFile = solutionFile;
                }

                //check for duplicate file names
                if (solutionFiles.Any(p => p.FileNameWithExtension == solutionFile.FileNameWithExtension))
                {
                    await
                        Response.WriteAsync(
                            Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest,
                                "duplicate file name")));
                    return;
                }

                solutionFiles.Add(solutionFile);
            }


            var solution = new CustomProjectSolution()
            {
                MainFileId =
                    null, //cannot be set when the solution is not saved yet (circular dependency ef error, see CreateFullExercise)
                PLangId = lastEditedPLang.Id,
                SolutionFiles = solutionFiles
            };

            #endregion

            #region full cusotm project and description

            //TODO currently we don't allow user assets!!
            customProjectFromFrontend.Description.Assets.Clear();

            var customProject = new CustomProject()
            {
                DisplayName = customProjectFromFrontend.DisplayName,
                LastEditorPLangId = lastEditedPLang.Id,
                Description = new CustomProjectDescription()
                {
                    Content = customProjectFromFrontend.Description.Content,
                    Assets = customProjectFromFrontend.Description.Assets.Select(p => new Asset()
                    {
                        DisplayName = p.DisplayName,
                        Content = p.Content.ToArray(),
                        MimeType = p.MimeType
                    }).ToList()
                },
                Tests = tests,
                Solutions = new List<CustomProjectSolution>() {solution},
                UserId = userId
            };

            #endregion


            using (var transaction = _context.Database.BeginTransaction())
            {
                try
                {
                    _context.CustomProjects.Add(customProject);
                    await _context.SaveChangesAsync();

                    //then set main file (else we get an error)

                    solution.MainFile = mainFile;

                    await _context.SaveChangesAsync();

                    transaction.Commit();
                }
                catch (Exception ex)
                {
                    transaction.Rollback();

                    await base.HandleDbError(ex);
                    return;
                }
            }

            await
                Response.WriteAsync(
                    Jc.Serialize(new BasicResponseWithData<int>(ResponseCode.Ok,
                        "", customProject.Id)));
        }

        [HttpPost("update/{customProjectId}/solution")]
        public async Task ChangeCustomProjectSolution(int customProjectId,
            [FromBody] CustomProjectSolutionFullBase solutionFullBase)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();

            var customProject = await _context.CustomProjects
                    .FirstOrDefaultAsync(p => p.UserId == userId && //just ensure the project is from the current user
                                              p.Id == customProjectId)
                ;

            if (customProject == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "custom project not found")));
                return;
            }

            var savedSolution = await _context.CustomProjectSolutions
                    .Where(p => p.CustomProjectId == customProjectId
                                && p.Id == solutionFullBase.Id)
                    .Include(p => p.SolutionFiles)
                    .Include(p => p.TestResults)
                    .FirstOrDefaultAsync()
                ;

            if (savedSolution == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "solution not found")));
                return;
            }

            if (solutionFullBase.SolutionFiles.Any(p => p.Id == solutionFullBase.MainFileId))
            {
                //ok
            }
            else
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "could not find main file")));
                return;
            }


            CustomProjectSolutionFile mainFile = null;
            bool mainFileDeleted = false;


            //check files to delete
            for (int i = 0; i < savedSolution.SolutionFiles.Count; i++)
            {
                var savedFile = savedSolution.SolutionFiles[i];

                if (solutionFullBase.SolutionFiles.Any(p => p.Id == savedFile.Id))
                {
                    //old file still there
                }
                else
                {
                    if (savedFile.Id == savedSolution.MainFileId)
                    {
                        mainFileDeleted = true;
                    }

                    //delete old file
                    savedSolution.SolutionFiles.Remove(savedFile);
                    i--;
                }
            }

            //add files
            int newCount = -1;
            foreach (var solutionFile in solutionFullBase.SolutionFiles)
            {
                var oldFile = savedSolution.SolutionFiles.FirstOrDefault(p => p.Id == solutionFile.Id);

                if (oldFile == null)
                {
                    //create the new file

                    var newFile = new CustomProjectSolutionFile()
                    {
                        // Id = newCount--, //ef now uses the negative id rather than generating auto incremented??, don't set and we get the db auto generated id after save
                        Content = solutionFile.Content,
                        DisplayIndex = solutionFile.DisplayIndex,
                        FileNameWithExtension = solutionFile.FileNameWithExtension,
                        IsDisplayed = solutionFile.UiIsDisplayed
                    };

                    if (solutionFullBase.MainFileId == solutionFile.Id)
                    {
                        mainFile = newFile;
                    }

                    savedSolution.SolutionFiles.Add(newFile);
                }
                else
                {
                    //update old file
                    oldFile.Content = solutionFile.Content;
                    oldFile.FileNameWithExtension = solutionFile.FileNameWithExtension;
                    oldFile.DisplayIndex = solutionFile.DisplayIndex;
                    oldFile.IsDisplayed = solutionFile.UiIsDisplayed;
                }
            }

            if (mainFileDeleted && mainFile == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest, "no new main file specified")));
                return;
            }

            foreach (var savedSolutionSolutionFile in savedSolution.SolutionFiles)
            {
                //check for duplicate file names
                if (savedSolution.SolutionFiles.Any(p => p.Id != savedSolutionSolutionFile.Id
                                                         && p.FileNameWithExtension == savedSolutionSolutionFile
                                                             .FileNameWithExtension))
                {
                    await
                        Response.WriteAsync(
                            Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest,
                                "duplicate file name")));
                    return;
                }
            }


            savedSolution.TestResults.Clear();


            customProject.LastEditorPLangId = savedSolution.PLangId;
            customProject.LastUpdatedAt = DateTimeHelper.GetUtcNow();

            using (var transaction = _context.Database.BeginTransaction())
            {
                try
                {
                    await _context.SaveChangesAsync();
                    if (mainFile != null)
                    {
                        //main file changed
                        savedSolution.MainFile = mainFile;
                        await _context.SaveChangesAsync();
                    }

                    transaction.Commit();
                }
                catch (Exception ex)
                {
                    transaction.Rollback();
                    await base.HandleDbError(ex);
                    return;
                }
            }

            var response = new CustomProjectSolutionFullBase()
            {
                Id = savedSolution.Id,
                CreatedAt = savedSolution.CreatedAt,
                MainFileId = savedSolution.MainFileId ?? 0,
                PLangId = savedSolution.PLangId,
                SolutionFiles = savedSolution.SolutionFiles.Select(p => new CustomProjectSolutionFileFullBase()
                {
                    Id = p.Id,
                    Content = p.Content,
                    DisplayIndex = p.DisplayIndex,
                    FileNameWithExtension = p.FileNameWithExtension,
                    UiIsDisplayed = p.IsDisplayed
                }).ToList()
            };

            await
                Response.WriteAsync(
                    Jc.Serialize(new BasicResponseWithData<CustomProjectSolutionFullBase>(ResponseCode.Ok,
                        "",
                        response))
                );
        }


        [HttpPost("update/{customProjectId}/description")]
        public async Task ChangeCustomProjectDescription(int customProjectId,
            [FromBody] CustomProjectDescriptionFullBase descriptionFullBase)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();

            var customProject = await _context.CustomProjects
                    .Where(p => p.UserId == userId && //just ensure the project is from the current user
                                p.Id == customProjectId)
                    .Include(p => p.Description)
                    .ThenInclude(p => p.Assets)
                    .FirstOrDefaultAsync()
                ;

            if (customProject == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "custom project not found")));
                return;
            }


            if (customProject.Description == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "custom project description not found")));
                return;
            }

            var oldDescription = customProject.Description;

            //assets are managed separately

            oldDescription.Content = descriptionFullBase.Content;

            customProject.LastUpdatedAt = DateTimeHelper.GetUtcNow();

            try
            {
                await _context.SaveChangesAsync();
            }
            catch (Exception ex)
            {
                await base.HandleDbError(ex);
                return;
            }

            await
                Response.WriteAsync(
                    Jc.Serialize(new BasicResponse(ResponseCode.Ok, "")));
        }


        [HttpPost("update/{customProjectId}/tests/create")]
        public async Task CreateCustomProjectTest(int customProjectId,
            [FromBody] CustomProjectTestFullBase testFullBase)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();

            var customProject = await _context.CustomProjects
                    .Include(p => p.Tests)
                    .FirstOrDefaultAsync(p => p.UserId == userId && //just ensure the project is from the current user
                                              p.Id == customProjectId)
                ;

            if (customProject == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "custom project not found")));
                return;
            }

            var testType = await _context.TestTypes.FirstOrDefaultAsync(p => p.Id == testFullBase.TestTypeId);

            if (testType == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "could not find test type")));
                return;
            }

            if (testFullBase.TestSettings == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest, "no test settings")));
                return;
            }

            var systemSetting = await _context.SystemSettings.FirstOrDefaultAsync();

            if (systemSetting == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest, "no system settings found")));
                return;
            }

            //systemSetting.MaxCustomTestsPerParticipation < 0 is unlimited
            if (systemSetting.MaxCustomTestsPerParticipation == 0 ||
                (systemSetting.MaxCustomTestsPerParticipation > 0 &&
                 customProject.Tests.Count + 1 > systemSetting.MaxCustomTestsPerParticipation))
            {
                //not allowed
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest,
                            "too much custom tests, max allowed is " + systemSetting.MaxCustomTestsPerParticipation)));
                return;
            }

            //assets are managed separately

            var newTest = new CustomProjectTest()
            {
                Content = testFullBase.Content,
                DisplayIndex = testFullBase.DisplayIndex,
                DisplayName = testFullBase.DisplayName.Trim(),
                TestType = testType,
                Weight = testFullBase.Weight,
                //CustomProjectId =  customProjectId //TODO needed?
                TestSettings = new TestSettings()
                {
                    MaxDiskSpaceInKb = testFullBase.TestSettings.MaxDiskSpaceInKb,
                    MemoryLimitInKb = testFullBase.TestSettings.MemoryLimitInKb,
                    TimeoutInMs = testFullBase.TestSettings.TimeoutInMs,
                    CompileTimeoutInMs = testFullBase.TestSettings.CompileTimeoutInMs,
                    CompilerOptions = testFullBase.TestSettings.CompilerOptions,
                }
            };

            customProject.Tests.Add(newTest);

            customProject.LastUpdatedAt = DateTimeHelper.GetUtcNow();


            try
            {
                await _context.SaveChangesAsync();
            }
            catch (Exception ex)
            {
                await base.HandleDbError(ex);
                return;
            }

            var response = new CustomProjectTestFullBase()
            {
                Id = newTest.Id,
                DisplayName = newTest.DisplayName,
                Content = newTest.Content,
                DisplayIndex = newTest.DisplayIndex,
                TestTypeId = newTest.TestTypeId,
                Weight = newTest.Weight,
                Files = new List<FilePreviewFromBackend>(),
                TestSettings = new TestSettingsFullBase()
                {
                    Id = newTest.TestSettings.Id,
                    MaxDiskSpaceInKb = newTest.TestSettings.MaxDiskSpaceInKb,
                    MemoryLimitInKb = newTest.TestSettings.MemoryLimitInKb,
                    TimeoutInMs = newTest.TestSettings.TimeoutInMs,
                    CompileTimeoutInMs = testFullBase.TestSettings.CompileTimeoutInMs,
                    CompilerOptions = testFullBase.TestSettings.CompilerOptions,
                }
            };

            await
                Response.WriteAsync(
                    Jc.Serialize(new BasicResponseWithData<CustomProjectTestFullBase>(ResponseCode.Ok, "", response)));
        }

        [HttpPost("update/{customProjectId}/tests/update")]
        public async Task ChangeCustomProjectTest(int customProjectId,
            [FromBody] CustomProjectTestFullBase testFullBase)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();

            var customProject = await _context.CustomProjects
                    .FirstOrDefaultAsync(p => p.UserId == userId && //just ensure the project is from the current user
                                              p.Id == customProjectId)
                ;

            if (customProject == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "custom project not found")));
                return;
            }

            var testType = await _context.TestTypes.FirstOrDefaultAsync(p => p.Id == testFullBase.TestTypeId);

            if (testType == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "could not find test type")));
                return;
            }

            if (testFullBase.TestSettings == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest, "no test settings")));
                return;
            }

            var oldTest = await _context.CustomProjectTests
                    .Where(p =>
                        p.Id == testFullBase.Id && p.CustomProjectId == customProjectId)
                    .Include(p => p.TestSettings)
                    .Include(p => p.TestResults)
                    .Include(p => p.AssetReferences)
                    .ThenInclude(p => p.FileReferenceUserFileAsset)
                    .FirstOrDefaultAsync()
                ;

            if (oldTest == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "test not found")));
                return;
            }

            oldTest.Content = testFullBase.Content;
            oldTest.DisplayIndex = testFullBase.DisplayIndex;
            oldTest.DisplayName = testFullBase.DisplayName;
            oldTest.Weight = testFullBase.Weight;

            oldTest.TestType = testType;

            oldTest.TestSettings.MaxDiskSpaceInKb = testFullBase.TestSettings.MaxDiskSpaceInKb;
            oldTest.TestSettings.MemoryLimitInKb = testFullBase.TestSettings.MemoryLimitInKb;
            oldTest.TestSettings.TimeoutInMs = testFullBase.TestSettings.TimeoutInMs;
            oldTest.TestSettings.CompilerOptions = testFullBase.TestSettings.CompilerOptions;

            //assets are managed separately

            //clear test results
            oldTest.TestResults.Clear();
            customProject.LastUpdatedAt = DateTimeHelper.GetUtcNow();

            try
            {
                await _context.SaveChangesAsync();
            }
            catch (Exception ex)
            {
                await base.HandleDbError(ex);
                return;
            }

            var response = new CustomProjectTestFullBase()
            {
                Id = oldTest.Id,
                DisplayName = oldTest.DisplayName,
                Content = oldTest.Content,
                DisplayIndex = oldTest.DisplayIndex,
                TestTypeId = oldTest.TestTypeId,
                Weight = oldTest.Weight,
                Files = oldTest.AssetReferences.Select(p => new FilePreviewFromBackend()
                {
                    //user cannot change test assets/or view them so not include data
                    Id = p.FileReferenceUserFileAsset.Id,
                    OriginalName = p.FileReferenceUserFileAsset.OriginalName,
                    MimeType = p.FileReferenceUserFileAsset.MimeType,
                    SizeInBytes = p.FileReferenceUserFileAsset.SizeInBytes,
                }).ToList(),
                TestSettings = new TestSettingsFullBase()
                {
                    Id = oldTest.TestSettings.Id,
                    MaxDiskSpaceInKb = oldTest.TestSettings.MaxDiskSpaceInKb,
                    MemoryLimitInKb = oldTest.TestSettings.MemoryLimitInKb,
                    TimeoutInMs = oldTest.TestSettings.TimeoutInMs,
                    CompileTimeoutInMs = testFullBase.TestSettings.CompileTimeoutInMs,
                    CompilerOptions = testFullBase.TestSettings.CompilerOptions,
                }
            };

            await
                Response.WriteAsync(
                    Jc.Serialize(new BasicResponseWithData<CustomProjectTestFullBase>(ResponseCode.Ok, "", response)));
        }

        [HttpDelete("update/{customProjectId}/tests/delete/{testId}")]
        public async Task DeleteCustomProjectTest(int customProjectId, int testId)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();

            var customProject = await _context.CustomProjects
                    .FirstOrDefaultAsync(p => p.UserId == userId && //just ensure the project is from the current user
                                              p.Id == customProjectId)
                ;

            if (customProject == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "custom project not found")));
                return;
            }

            var oldTest = await _context.CustomProjectTests
                    .FirstOrDefaultAsync(p => p.Id == testId && p.CustomProjectId == customProjectId)
                ;

            if (oldTest == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "test not found")));
                return;
            }

            _context.CustomProjectTests.Remove(oldTest);
            customProject.LastUpdatedAt = DateTimeHelper.GetUtcNow();

            try
            {
                await _context.SaveChangesAsync();
            }
            catch (Exception ex)
            {
                await base.HandleDbError(ex);
                return;
            }

            await Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.Ok, "")));
        }


        /// <summary>
        /// gets (downloads) the custom project test asset
        /// </summary>
        /// <param name="customProjectId">the custom project</param>
        /// <param name="testId">the test id</param>
        /// <param name="assetId">the asset id</param>
        [HttpGet("asset/{customProjectId}/{testId}/{assetId}")]
        public async Task GetCustomProjectTestAsset(int customProjectId, int testId, int assetId)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();


            var test = await _context.CustomProjectTests
                    .Include(p => p.TestAssets)
                    .Where(p => p.Id == testId && p.CustomProjectId == customProjectId &&
                                p.CustomProject.UserId == userId)
                    .Select(p => new
                    {
                        Asset = p.TestAssets.FirstOrDefault(_asset => _asset.Id == assetId)
                    })
                    .FirstOrDefaultAsync()
                ;


            if (test == null)
            {
                Response.StatusCode = 406;
                await
                    Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "test not found")));
                return;
            }

            if (test.Asset == null)
            {
                Response.StatusCode = 406;
                await
                    Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "asset not found")));
                return;
            }

            var fileName = "asset";
            if (Files.ValidFileName(test.Asset.DisplayName))
            {
                fileName = test.Asset.DisplayName;
            }
            else
            {
                //just try to get a file extension...

                var lastIndex = test.Asset.DisplayName.LastIndexOf(".", StringComparison.Ordinal);
                if (lastIndex == -1)
                {
                    //no file extension...
                    //TODO maybe get from mime type somehow?
                }
                else
                {
                    var extension = test.Asset.DisplayName.Substring(lastIndex);
                    fileName += extension;
                }
            }

            var asset = new TestAsset()
            {
                Id = test.Asset.Id,
                Content = test.Asset.Content,
                DisplayName = fileName,
                MimeType = test.Asset.MimeType,
            };


            Response.Headers.Add(new KeyValuePair<string, StringValues>("Content-Disposition",
                "attachment; filename=\"" + asset.DisplayName + "\""));

            byte[] content = asset.Content;


            Response.Body.Write(content, 0, content.Length);
            await Response.Body.FlushAsync();
        }

        [HttpPost("{customProjectId}/properties")]
        public async Task UpdateCustomProjectProperties(int customProjectId,
            [FromBody] CustomProjectPropertiesFromFrontend propertiesFromFrontend)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();

            var customProject = await _context.CustomProjects
                    .FirstOrDefaultAsync(p => p.UserId == userId && //just ensure the project is from the current user
                                              p.Id == customProjectId)
                ;

            if (customProject == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "custom project not found")));
                return;
            }

            customProject.DisplayName = propertiesFromFrontend.DisplayName;

            try
            {
                await _context.SaveChangesAsync();
            }
            catch (Exception ex)
            {
                await base.HandleDbError(ex);
                return;
            }

            await
                Response.WriteAsync(
                    Jc.Serialize(new BasicResponse(ResponseCode.Ok, "")));
        }


        //TODO func to change display index of tests

        #region custom project description/markdown assets

        /// <summary>
        /// this will upload (receive) the file and connect it to the custom project description
        /// </summary>
        /// <param name="customProjectId"></param>
        /// <param name="file"></param>
        /// <returns></returns>
        [HttpPost("upload/create/custom/description/{customProjectId}")]
        public async Task UploadExerciseDescriptionAssetFile(int customProjectId, IFormFile file)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();

            var customProject = await _context.CustomProjects
                    .Include(p => p.Description)
                    .FirstOrDefaultAsync(p => p.UserId == userId && //just ensure the project is from the current user
                                              p.Id == customProjectId)
                ;

            if (customProject == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "custom project not found")));
                return;
            }

            if (customProject.Description == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "custom project description not found")));
                return;
            }

            if (file == null || file.Length == 0)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "no files selected")));
                return;
            }

            var targetHardDrivePath = Files.GetUploadFilePath(UploadDirType.UserAssets);
            FileReferenceUserFileAsset fileRef = null;
            try
            {
                using (var transaction = await _context.Database.BeginTransactionAsync())
                {
                    try
                    {
                        fileRef = new FileReferenceUserFileAsset()
                        {
                            Hash = "",
                            MimeType = file.ContentType,
                            OriginalName = file.FileName,
                        };

                        var connection = new CustomProjectDescriptionWithFileAsAssetReference()
                        {
                            FileReferenceUserFileAsset = fileRef,
                            CustomProjectDescription = customProject.Description
                        };

                        _context.CustomProjectDescriptionWithFileAsAssetReferences.Add(connection);
                        await _context.SaveChangesAsync();

                        Tuple<string, long> _tuple =
                            await Files.CreateUploadedFile(targetHardDrivePath, fileRef.Id, file);
                        fileRef.Hash = _tuple.Item1;
                        fileRef.SizeInBytes = _tuple.Item2;

                        await _context.SaveChangesAsync();
                        transaction.Commit();
                    }
                    catch (Exception e)
                    {
                        transaction.Rollback();
                        await base.HandleDbError(e);
                        return;
                    }
                }
            }
            catch (Exception e)
            {
                await base.HandleDbError(e);
                return;
            }


            var assetPreviewForFrontend = new FilePreviewFromBackend()
            {
                Id = fileRef.Id,
                MimeType = fileRef.MimeType,
                OriginalName = fileRef.OriginalName,
                SizeInBytes = fileRef.SizeInBytes
            };


            await
                Response.WriteAsync(
                    Jc.Serialize(
                        new BasicResponseWithData<FilePreviewFromBackend>(ResponseCode.Ok, "",
                            assetPreviewForFrontend)));
        }

        /// <summary>
        /// adds an already uploaded file to the custom project description
        /// </summary>
        /// <param name="customProjectId"></param>
        /// <param name="fileReferenceId"></param>
        /// <returns></returns>
        [HttpPost("upload/attach/custom/description/{customProjectId}/{fileReferenceId}")]
        public async Task AttachExitingFileToExerciseDescription(int customProjectId, int fileReferenceId)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();

            var customProject = await _context.CustomProjects
                    .Include(p => p.Description)
                    .FirstOrDefaultAsync(p => p.UserId == userId && //just ensure the project is from the current user
                                              p.Id == customProjectId)
                ;

            if (customProject == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "custom project not found")));
                return;
            }

            if (customProject.Description == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "custom project description not found")));
                return;
            }

            var fileRef = await _context.FileReferenceUserFileAssets
                .FirstOrDefaultAsync(p => p.Id == fileReferenceId);

            if (fileRef == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "asset file not found")));
                return;
            }

            var connection = new CustomProjectDescriptionWithFileAsAssetReference()
            {
                FileReferenceUserFileAsset = fileRef,
                CustomProjectDescription = customProject.Description
            };

            try
            {
                _context.CustomProjectDescriptionWithFileAsAssetReferences.Add(connection);
                await _context.SaveChangesAsync();
            }
            catch (Exception e)
            {
                await base.HandleDbError(e);
                return;
            }

            //we don't need to send back information because the frontend need to load the file anyway and will
            //call GetSingleExerciseDescriptionAssetFile which includes all needed information

            await
                Response.WriteAsync(
                    Jc.Serialize(new BasicResponse(ResponseCode.Ok, "")));
        }

        ///  <summary>
        ///  removes the file from the custom project (custom project description)
        /// 
        ///  IF the file was only attached to this last custom project then the file gets deleted
        ///    (because there are no other connection to this file)
        ///  </summary>
        ///  <param name="customProjectId"></param>
        /// <param name="fileReferenceId"></param>
        /// <returns></returns>
        [HttpDelete("upload/detach/custom/description/{customProjectId}/{fileReferenceId}")]
        public async Task DetachOrDeleteFileFromExerciseDescription(int customProjectId, int fileReferenceId)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();

            var customProject = await _context.CustomProjects
                    .Include(p => p.Description)
                    .FirstOrDefaultAsync(p => p.UserId == userId && //just ensure the project is from the current user
                                              p.Id == customProjectId)
                ;

            if (customProject == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "custom project not found")));
                return;
            }

            if (customProject.Description == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "custom project description not found")));
                return;
            }

            var assetFileReference = await _context.CustomProjectDescriptionWithFileAsAssetReferences
                .FirstOrDefaultAsync(p =>
                    p.FileReferenceUserFileAssetId == fileReferenceId &&
                    p.CustomProjectDescriptionId == customProject.Description.Id);

            if (assetFileReference == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "file reference connection not found")));
                return;
            }

            //then check if this was the last reference to this file from an exercise...

            //then check if this was the last reference to this file from an exercise...
            //for user files we need to check:
            //  custom test assets
            //  custom project description assets
            //  custom project test assets
            //only if the sum is 1 we can delete it!

            var referenceCount = await _context.CustomProjectTestWithFileAsAssetReferences
                    .CountAsync(p => p.FileReferenceUserFileAssetId == assetFileReference.FileReferenceUserFileAssetId)
                ;

            var referenceCount2 = await _context.CustomProjectDescriptionWithFileAsAssetReferences
                    .CountAsync(p => p.FileReferenceUserFileAssetId == assetFileReference.FileReferenceUserFileAssetId)
                ;

            var referenceCount3 = await _context.CustomTestWithFileAsAssetReferences
                    .CountAsync(p => p.FileReferenceUserFileAssetId == assetFileReference.FileReferenceUserFileAssetId)
                ;

            if (referenceCount + referenceCount2 + referenceCount3 == 1) //this is the last reference
            {
                //no exercise has a reference to this file... delete


                var fileReference = await _context.FileReferenceUserFileAssets
                    .FirstOrDefaultAsync(p => p.Id == assetFileReference.FileReferenceUserFileAssetId);

                if (fileReference == null)
                {
                    await
                        Response.WriteAsync(
                            Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "file reference not found")));
                    return;
                }


                using (var transaction = await _context.Database.BeginTransactionAsync())
                {
                    try
                    {
                        _context.FileReferenceUserFileAssets.Remove(fileReference);
                        await _context.SaveChangesAsync();

                        var targetHardDrivePath = Files.GetUploadFilePath(UploadDirType.UserAssets);

                        Files.DeleteUploadedFile(targetHardDrivePath, fileReference.Id);

                        transaction.Commit();
                    }
                    catch (Exception e)
                    {
                        transaction.Rollback();
                        await base.HandleDbError(e);
                        return;
                    }
                }
            }
            else
            {
                try
                {
                    _context.CustomProjectDescriptionWithFileAsAssetReferences.Remove(assetFileReference);
                    await _context.SaveChangesAsync();
                }
                catch (Exception e)
                {
                    await base.HandleDbError(e);
                    return;
                }
            }


            await
                Response.WriteAsync(
                    Jc.Serialize(new BasicResponse(ResponseCode.Ok, "")));
        }

        /// <summary>
        /// this is needed when we attach an existing file to an custom project then we need to content in the frontend
        /// </summary>
        /// <param name="customProjectId"></param>
        /// <param name="fileReferenceId"></param>
        /// <returns></returns>
        [HttpGet("upload/get/description/content/{customProjectId}/{fileReferenceId}")]
        public async Task GetSingleExerciseDescriptionAssetFile(int customProjectId, int fileReferenceId)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();


            var assetForFrontend =
                await _GetSingleExerciseDescriptionAssetFile(userId, customProjectId, fileReferenceId, false);

            if (assetForFrontend == null) return;

            await
                Response.WriteAsync(
                    Jc.Serialize(
                        new BasicResponseWithData<FileWithData>(ResponseCode.Ok, "", assetForFrontend)));
        }


        internal async Task<FileWithData> _GetSingleExerciseDescriptionAssetFile(int userId, int customProjectId,
            int fileReferenceId, bool isFileDownload)
        {
            var customProject = await _context.CustomProjects
                    .Include(p => p.Description)
                    .FirstOrDefaultAsync(p => p.UserId == userId && //just ensure the project is from the current user
                                              p.Id == customProjectId)
                ;

            if (customProject == null)
            {
                if (isFileDownload)
                {
                    Response.StatusCode = 406;
                }

                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "custom project not found")));
                return null;
            }

            if (customProject.Description == null)
            {
                if (isFileDownload)
                {
                    Response.StatusCode = 406;
                }

                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "custom project description not found")));
                return null;
            }

            //file should only be read if it is connected to the exercise

            var connection = await _context.CustomProjectDescriptionWithFileAsAssetReferences
                    .Where(p => p.CustomProjectDescriptionId == customProject.Description.Id &&
                                p.FileReferenceUserFileAssetId == fileReferenceId)
                    .AsNoTracking()
                    .FirstOrDefaultAsync()
                ;

            if (connection == null)
            {
                if (isFileDownload)
                {
                    Response.StatusCode = 406;
                }

                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound,
                            "file is not connected to the custom project description")));
                return null;
            }

            var fileRef = await _context.FileReferenceUserFileAssets.FirstOrDefaultAsync(p => p.Id == fileReferenceId);

            if (fileRef == null)
            {
                if (isFileDownload)
                {
                    Response.StatusCode = 406;
                }

                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "file was not found")));
                return null;
            }

            var basePath = Files.GetUploadFilePath(UploadDirType.UserAssets);

            var filesDict = await Files.ReadUploadedFiles(basePath, fileReferenceId);


            var assetForFrontend = new FileWithData()
            {
                Id = fileRef.Id,
                DisplayName = fileRef.OriginalName,
                MimeType = fileRef.MimeType,
                Content = filesDict[fileReferenceId]
            };

            return assetForFrontend;
        }

        #endregion

        #region custom project custom test assets

        /// <summary>
        /// this will upload (receive) the file and connect it to the custom project custom test
        /// </summary>
        /// <param name="customProjectId"></param>
        /// <param name="customTestId"></param>
        /// <param name="file"></param>
        /// <returns></returns>
        [HttpPost("upload/create/custom/tests/{customProjectId}/{customTestId}")]
        public async Task UploadCustomProjectCustomTestAssetFile(int customProjectId, int customTestId, IFormFile file)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();

            var customProject = await _context.CustomProjects
                    .FirstOrDefaultAsync(p => p.UserId == userId && //just ensure the project is from the current user
                                              p.Id == customProjectId)
                ;

            if (customProject == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "custom project not found")));
                return;
            }


            var oldTest = await _context.CustomProjectTests
                    .Where(p =>
                        p.Id == customTestId && p.CustomProjectId == customProjectId)
                    .FirstOrDefaultAsync()
                ;

            if (oldTest == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "test not found")));
                return;
            }

            if (file == null || file.Length == 0)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "no files selected")));
                return;
            }


            var targetHardDrivePath = Files.GetUploadFilePath(UploadDirType.UserAssets);
            FileReferenceUserFileAsset fileRef = null;
            try
            {
                using (var transaction = await _context.Database.BeginTransactionAsync())
                {
                    try
                    {
                        fileRef = new FileReferenceUserFileAsset()
                        {
                            Hash = "",
                            MimeType = file.ContentType,
                            OriginalName = file.FileName,
                        };

                        var connection = new CustomProjectTestWithFileAsAssetReference()
                        {
                            FileReferenceUserFileAsset = fileRef,
                            CustomProjectTest = oldTest
                        };

                        _context.CustomProjectTestWithFileAsAssetReferences.Add(connection);
                        await _context.SaveChangesAsync();

                        Tuple<string, long> _tuple =
                            await Files.CreateUploadedFile(targetHardDrivePath, fileRef.Id, file);
                        fileRef.Hash = _tuple.Item1;
                        fileRef.SizeInBytes = _tuple.Item2;

                        await _context.SaveChangesAsync();
                        transaction.Commit();
                    }
                    catch (Exception e)
                    {
                        transaction.Rollback();
                        await base.HandleDbError(e);
                        return;
                    }
                }
            }
            catch (Exception e)
            {
                await base.HandleDbError(e);
                return;
            }


            var assetPreviewForFrontend = new FilePreviewFromBackend()
            {
                Id = fileRef.Id,
                MimeType = fileRef.MimeType,
                OriginalName = fileRef.OriginalName,
                SizeInBytes = fileRef.SizeInBytes
            };


            await
                Response.WriteAsync(
                    Jc.Serialize(
                        new BasicResponseWithData<FilePreviewFromBackend>(ResponseCode.Ok, "",
                            assetPreviewForFrontend)));
        }

        /// <summary>
        /// adds an already uploaded file to the custom project custom test
        /// </summary>
        /// <param name="customProjectId"></param>
        /// <param name="customTestId"></param>
        /// <param name="fileReferenceId"></param>
        /// <returns></returns>
        [HttpPost("upload/attach/custom/tests/{customProjectId}/{customTestId}/{fileReferenceId}")]
        public async Task AttachExitingFileToCustomProjectCustomTest(int customProjectId, int customTestId,
            int fileReferenceId)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();

            var customProject = await _context.CustomProjects
                    .FirstOrDefaultAsync(p => p.UserId == userId && //just ensure the project is from the current user
                                              p.Id == customProjectId)
                ;

            if (customProject == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "custom project not found")));
                return;
            }

            var oldTest = await _context.CustomProjectTests
                    .Where(p =>
                        p.Id == customTestId && p.CustomProjectId == customProjectId)
                    .FirstOrDefaultAsync()
                ;

            if (oldTest == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "test not found")));
                return;
            }


            var fileRef = await _context.FileReferenceUserFileAssets
                .FirstOrDefaultAsync(p => p.Id == fileReferenceId);

            if (fileRef == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "asset file not found")));
                return;
            }

            var connection = new CustomProjectTestWithFileAsAssetReference()
            {
                FileReferenceUserFileAsset = fileRef,
                CustomProjectTest = oldTest
            };

            try
            {
                _context.CustomProjectTestWithFileAsAssetReferences.Add(connection);
                await _context.SaveChangesAsync();
            }
            catch (Exception e)
            {
                await base.HandleDbError(e);
                return;
            }

            //we need to send some information so the frontend can display the asset file

            var filePreview = new FilePreviewFromBackend()
            {
                Id = fileRef.Id,
                OriginalName = fileRef.OriginalName,
                SizeInBytes = fileRef.SizeInBytes,
                MimeType = fileRef.MimeType
            };


            await
                Response.WriteAsync(
                    Jc.Serialize(new BasicResponseWithData<FilePreviewFromBackend>(ResponseCode.Ok, "", filePreview)));
        }


        ///  <summary>
        ///  removes the file from the custom test 
        /// 
        ///  IF the file was only attached to this last test then the file gets deleted
        ///    (because there are no other connection to this file)
        ///  </summary>
        ///  <param name="customProjectId"></param>
        /// <param name="customTestId"></param>
        /// <param name="fileReferenceId"></param>
        /// <returns></returns>
        [HttpDelete("upload/detach/custom/tests/{customProjectId}/{customTestId}/{fileReferenceId}")]
        public async Task DetachOrDeleteFileFromCustomProjectCustomTest(int customProjectId, int customTestId,
            int fileReferenceId)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();

            var customProject = await _context.CustomProjects
                    .FirstOrDefaultAsync(p => p.UserId == userId && //just ensure the project is from the current user
                                              p.Id == customProjectId)
                ;

            if (customProject == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "custom project not found")));
                return;
            }


            var oldTest = await _context.CustomProjectTests
                    .Where(p =>
                        p.Id == customTestId && p.CustomProjectId == customProjectId)
                    .FirstOrDefaultAsync()
                ;

            if (oldTest == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "test not found")));
                return;
            }

            var assetFileReference = await _context.CustomProjectTestWithFileAsAssetReferences
                .FirstOrDefaultAsync(p =>
                    p.FileReferenceUserFileAssetId == fileReferenceId && p.CustomProjectTestId == oldTest.Id);

            if (assetFileReference == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "file reference connection not found")));
                return;
            }

            //then check if this was the last reference to this file from an exercise...
            //for user files we need to check:
            //  custom test assets
            //  custom project description assets
            //  custom project test assets
            //only if the sum is 1 we can delete it!

            var referenceCount = await _context.CustomProjectTestWithFileAsAssetReferences
                    .CountAsync(p => p.FileReferenceUserFileAssetId == assetFileReference.FileReferenceUserFileAssetId)
                ;

            var referenceCount2 = await _context.CustomProjectDescriptionWithFileAsAssetReferences
                    .CountAsync(p => p.FileReferenceUserFileAssetId == assetFileReference.FileReferenceUserFileAssetId)
                ;

            var referenceCount3 = await _context.CustomTestWithFileAsAssetReferences
                    .CountAsync(p => p.FileReferenceUserFileAssetId == assetFileReference.FileReferenceUserFileAssetId)
                ;

            if (referenceCount + referenceCount2 + referenceCount3 == 1) //this is the last reference
            {
                //no test has a reference to this file... delete


                var fileReference = await _context.FileReferenceUserFileAssets
                    .FirstOrDefaultAsync(p => p.Id == assetFileReference.FileReferenceUserFileAssetId);


                if (fileReference == null)
                {
                    await
                        Response.WriteAsync(
                            Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "file reference not found")));
                    return;
                }

                using (var transaction = await _context.Database.BeginTransactionAsync())
                {
                    try
                    {
                        _context.FileReferenceUserFileAssets.Remove(fileReference);
                        await _context.SaveChangesAsync();

                        var targetHardDrivePath = Files.GetUploadFilePath(UploadDirType.UserAssets);

                        Files.DeleteUploadedFile(targetHardDrivePath, fileReference.Id);

                        transaction.Commit();
                    }
                    catch (Exception e)
                    {
                        transaction.Rollback();
                        await base.HandleDbError(e);
                        return;
                    }
                }
            }
            else
            {
                try
                {
                    _context.CustomProjectTestWithFileAsAssetReferences.Remove(assetFileReference);
                    await _context.SaveChangesAsync();
                }
                catch (Exception e)
                {
                    await base.HandleDbError(e);
                    return;
                }
            }


            await
                Response.WriteAsync(
                    Jc.Serialize(new BasicResponse(ResponseCode.Ok, "")));
        }

        #endregion
    }


    public class CustomProjectPropertiesFromFrontend
    {
        /// <summary>
        /// the name
        /// </summary>
        public string DisplayName { get; set; }
    }

    public class CustomProjectFullBase
    {
        public int Id { get; set; }

        /// <summary>
        /// the name
        /// </summary>
        public string DisplayName { get; set; }

        /// <summary>
        /// the last edited plang... set through the custom solution
        /// </summary>
        public int LastEditorPLangId { get; set; }

        /// <summary>
        /// some description
        /// </summary>
        public CustomProjectDescriptionFullBase Description { get; set; }

        /// <summary>
        /// the solution
        /// </summary>
        public CustomProjectSolutionFullBase Solution { get; set; }

        /// <summary>
        /// the tests
        /// </summary>
        public List<CustomProjectTestFullBase> Tests { get; set; }
    }

    /// <summary>
    /// the tests include the test asset content
    /// </summary>
    public class CustomProjectWithData
    {
        public int Id { get; set; }

        /// <summary>
        /// the name
        /// </summary>
        public string DisplayName { get; set; }

        /// <summary>
        /// the last edited plang... set through the custom solution
        /// </summary>
        public int LastEditorPLangId { get; set; }

        /// <summary>
        /// some description
        /// </summary>
        public CustomProjectDescriptionFullBase Description { get; set; }

        /// <summary>
        /// the solution
        /// </summary>
        public CustomProjectSolutionFullBase Solution { get; set; }

        /// <summary>
        /// the tests
        /// </summary>
        public List<CustomProjectTestWithData> Tests { get; set; }
    }


    public class CustomProjectDescriptionFullBase
    {
        public CustomProjectDescriptionFullBase()
        {
            Assets = new List<FileWithData>();
        }

        public int Id { get; set; }

        /// <summary>
        /// the content
        /// </summary>
        public string Content { get; set; }

        /// <summary>
        /// the contents are not included by default
        /// </summary>
        public List<FileWithData> Assets { get; set; }
    }


    public class CustomProjectSolutionFullBase
    {
        public CustomProjectSolutionFullBase()
        {
            SolutionFiles = new List<CustomProjectSolutionFileFullBase>();
        }

        public int Id { get; set; }

        public int PLangId { get; set; }

        /// <summary>
        /// the datetime where the version was created
        /// </summary>
        public DateTime? CreatedAt { get; set; }

        public int MainFileId { get; set; }

        /// <summary>
        /// the connected files for this solution
        /// </summary>
        public List<CustomProjectSolutionFileFullBase> SolutionFiles { get; set; }

        //don't include the tests results here so that we can use this class for frontend and backend
        //e.g. when saving a new solution version we don't need this...
        //public List<CustomProjectTestWithSolutionAsTestResult> TestResults { get; set; }
    }

    /// <summary>
    /// should be the same as ExerciseCodeTemplateFileFullBase... so that the frontend only need to handle on
    /// file type
    /// </summary>
    public class CustomProjectSolutionFileFullBase : ISolutionFile
    {
        public int Id { get; set; }

        /// <summary>
        /// the file nname
        /// </summary>
        public string FileNameWithExtension { get; set; }

        /// <summary>
        /// the content of the file
        /// </summary>
        public string Content { get; set; }

        /// <summary>
        /// ui property, true: is displayed as tab, false: not (only displayed in project explorer)
        /// </summary>
        public bool UiIsDisplayed { get; set; }

        /// <summary>
        /// the order index how the file is displayed in the editor
        /// NOT USED YET
        /// </summary>
        public int DisplayIndex { get; set; }
    }

    public class CustomProjectTestFullBase
    {
        public CustomProjectTestFullBase()
        {
            Files = new List<FilePreviewFromBackend>();
        }

        public int Id { get; set; }

        /// <summary>
        /// the name
        /// </summary>
        public string DisplayName { get; set; }

        /// <summary>
        /// the content for the ... see the black-box-test protocol
        /// </summary>
        public string Content { get; set; }

        /// <summary>
        /// the weight 
        /// </summary>
        public int Weight { get; set; }

        /// <summary>
        /// the index for the frontend sorting (>= 0, 0 is first, 1 second, ...)
        /// </summary>
        public int DisplayIndex { get; set; }

        public int TestTypeId { get; set; }

        /// <summary>
        /// assets
        /// </summary>
        public List<FilePreviewFromBackend> Files { get; set; }

        /// <summary>
        /// the test settings
        /// </summary>
        public TestSettingsFullBase TestSettings { get; set; }
    }

    public class CustomProjectTestWithData
    {
        public CustomProjectTestWithData()
        {
            Files = new List<TestAssetFromBackendWithData>();
        }

        public int Id { get; set; }

        /// <summary>
        /// the name
        /// </summary>
        public string DisplayName { get; set; }

        /// <summary>
        /// the content for the ... see the black-box-test protocol
        /// </summary>
        public string Content { get; set; }

        /// <summary>
        /// the weight 
        /// </summary>
        public int Weight { get; set; }

        /// <summary>
        /// the index for the frontend sorting (>= 0, 0 is first, 1 second, ...)
        /// </summary>
        public int DisplayIndex { get; set; }

        public int TestTypeId { get; set; }

        /// <summary>
        /// assets
        /// </summary>
        public List<TestAssetFromBackendWithData> Files { get; set; }

        /// <summary>
        /// the test settings
        /// </summary>
        public TestSettingsFullBase TestSettings { get; set; }
    }
}
